@@ -36,26 +36,6 @@
 #include "lr_parser.h"
 #include "ui.h"
 
-char *
-gc_strdup(char *str)
-{
-    char *dup;
-    int len;
-
-    if (!str) {
-        return NULL;
-    }
-
-    len = strlen(str);
-    dup = malloc(len+1);
-    assert(dup);
-    memset(dup, 0, len+1);
-    strncpy(dup, str, len);
-
-    return(dup);
-}
-
-
 void
 make_exp(char *number, int t[MP_SIZE])
 {
@@ -67,7 +47,7 @@ make_exp(char *number, int t[MP_SIZE])
     int MP_b[MP_SIZE];
 
     assert(number);
-    a = gc_strdup(number);
+    a = strdup(number);
     assert(a);
 
     for (i = 0; !((a[i] == 'e') || (a[i] == 'E')); i++) {
@@ -91,37 +71,10 @@ make_exp(char *number, int t[MP_SIZE])
 }
 
 
-static void
-update_undo_redo_button_sensitivity(void)
-{
-    int undo = 0;
-    int redo = 0;
-
-    if (v->h.current != v->h.end) {
-        redo = 1;
-    }
-
-    if (v->h.current != v->h.begin) {
-        undo = 1;
-    }
-
-    ui_set_undo_enabled(undo, redo);
-}
-
-
 void
 clear_undo_history(void)
 {
-    int i = v->h.begin;
-    while (i != v->h.end) {
-        if (i != v->h.current) {
-            free(v->h.e[i].expression);
-            v->h.e[i].expression = NULL;
-        }
-        i = ((i + 1) % UNDO_HISTORY_LENGTH);
-    }
-    v->h.begin = v->h.end = v->h.current;
-    update_undo_redo_button_sensitivity();
+    display_clear_stack();
 }
 
 
@@ -132,77 +85,12 @@ get_state(void)
 }
 
 
-static void
-copy_state(struct exprm_state *dst, struct exprm_state *src)
-{
-    MEMCPY(dst, src, sizeof(struct exprm_state));
-    dst->expression = gc_strdup(src->expression);
-}
-
-
-static void
-purge_redo_history(void)
-{
-    if (v->h.current != v->h.end) {
-        int i = v->h.current;
-
-        do {
-            i = ((i + 1) % UNDO_HISTORY_LENGTH);
-            free(v->h.e[i].expression);
-            v->h.e[i].expression = gc_strdup("Ans");
-        } while (i != v->h.end);
-    }
-
-    v->h.end = v->h.current;
-}
-
-
-static void
-new_state(void)
-{
-    int c;
-
-    purge_redo_history();
-
-    c = v->h.current;
-    v->h.end = v->h.current = ((v->h.current + 1) % UNDO_HISTORY_LENGTH);
-    if (v->h.current == v->h.begin) {
-        free(v->h.e[v->h.begin].expression);
-        v->h.e[v->h.begin].expression = NULL;
-        v->h.begin = ((v->h.begin + 1) % UNDO_HISTORY_LENGTH);
-    }
-
-    copy_state(&(v->h.e[v->h.current]), &(v->h.e[c]));
-    update_undo_redo_button_sensitivity();
-}
-
-
 void
 perform_undo(void)
 {
-    struct exprm_state *e;
-    
-    if (v->h.current != v->h.begin) {
-        v->h.current = ((v->h.current - 1) % UNDO_HISTORY_LENGTH);
-        ui_set_statusbar("", "");
-    } else {
-        ui_set_statusbar(_("No undo history"), "gtk-dialog-warning");
-    }
-    update_undo_redo_button_sensitivity();
-    
-    e = get_state();
-    refresh_display(e->cursor);
+    display_pop();
 }
 
-struct exprm_state *
-peek_previous_state(void)
-{
-    if (v->h.current != v->h.begin) {
-        return &(v->h.e[(v->h.current - 1) % UNDO_HISTORY_LENGTH]);
-    } else {
-        return NULL;
-    }
-}
 
 static int
 is_undo_step(void)
@@ -214,13 +102,7 @@ is_undo_step(void)
 void
 perform_redo(void)
 {
-    if (v->h.current != v->h.end) {
-        v->h.current = ((v->h.current + 1) % UNDO_HISTORY_LENGTH);
-        ui_set_statusbar("", "");
-    } else {
-        ui_set_statusbar(_("No redo steps"), "gtk-dialog-warning");
-    }
-    update_undo_redo_button_sensitivity();
+    display_unpop();
 }
 
 
@@ -258,203 +140,21 @@ do_business(void)     /* Perform special business mode calculations. */
     } else if (v->current == KEY_FINC_TERM) {
         calc_term(v->MPdisp_val);
     }
-    show_display(v->MPdisp_val);
-}
-
-
-int
-exp_insert(char *text, int cursor)
-{
-    char buf[MAXLINE], *display;
-    struct exprm_state *e = get_state();
-    
-    if (cursor < 0) {
-        SNPRINTF(buf, MAXLINE, "%s%s", e->expression, text);
-    } else {
-        display = ui_get_display();
-        SNPRINTF(buf, MAXLINE, "%.*s%s%s", cursor, display, text, display + cursor);
-        cursor += strlen(text);
-    }
-    exp_replace(buf);
-    
-    return cursor;
-}
-
-
-void
-exp_clear(void)
-{
-    exp_replace("");
-}
-
-
-int
-usable_num(int MPnum[MP_SIZE])
-{
-    struct exprm_state *e = get_state();
-    if (e->expression[0] == '\0') {
-        return ce_parse("0", MPnum);
-    } else {
-        return ce_parse(e->expression, MPnum);
-    }
-}
-
-
-static int
-exp_has_postfix(char *str, char *postfix)
-{
-    int len, plen;
-
-    if (!str) {
-        return(0);
-    }
-
-    assert(postfix);
-
-    len = strlen(str);
-    plen = strlen(postfix);
-
-    if (plen > len) {
-        return(0);
-    }
-
-    if (!strcasecmp(str + len - plen, postfix)) {
-        return(1);
-    } else {
-        return(0);
-    }
-}
-
-
-static int
-exp_backspace(int cursor)
-{
-    char buf[MAXLINE] = "", *display;
-    struct exprm_state *e = get_state();
-    int i, MP_reg[MP_SIZE];
-
-    /* If cursor is at end of the line then delete the last character preserving accuracy */
-    if (cursor < 0) {
-        if (exp_has_postfix(e->expression, "Ans")) {
-            char *ans = make_number(e->ans, v->base, FALSE);
-            char *t;
-            t = str_replace(e->expression, "Ans", ans);
-            free(e->expression);
-            e->expression = t;
-        } else {
-            for (i = 0; i < 10; i++) {
-                SNPRINTF(buf, MAXLINE, "R%d", i);
-                if (exp_has_postfix(e->expression, buf)) {
-                    do_rcl_reg(i, MP_reg);
-                    display = make_number(MP_reg, v->base, FALSE);
-                    /* Remove "Rx" postfix and replace with backspaced number */
-                    SNPRINTF(buf, MAXLINE, "%.*s%s", strlen(e->expression) - 2, e->expression - 3, display);
-                    exp_replace(buf);
-                    return cursor - 1;
-                }
-            }
-        }
-
-        SNPRINTF(buf, MAXLINE, "%.*s", strlen(e->expression) - 1, e->expression);
-    } else if (cursor > 0) {
-        display = ui_get_display();
-        SNPRINTF(buf, MAXLINE, "%.*s%s", cursor - 1, display, display + cursor);
-    } else {
-        return cursor; /* At the start of the line */
-    }
-
-    exp_replace(buf);
-    return cursor - 1;
-}
-
-
-static int
-exp_delete(int cursor)
-{
-    char buf[MAXLINE] = "", *display;
-    
-    if (cursor >= 0) {
-        display = ui_get_display();
-        SNPRINTF(buf, MAXLINE, "%.*s%s", cursor, display, display + cursor + 1);
-        exp_replace(buf);
-    }
-
-    return cursor;
-}
-
-
-void
-exp_replace(char *text)
-{
-    struct exprm_state *e;
-    e = get_state();
-    if (e->expression != NULL) {
-        free(e->expression);
-    }
-    e->expression = gc_strdup(text);
+    display_set_number(v->MPdisp_val);
 }
 
 
 static void
 exp_negate(void)
 {
-    struct exprm_state *e = get_state();
-
-    /* Ending zero + parenthesis + minus */
-    int len = strlen(e->expression) + 4;
-    char *exp = malloc(len);
-
-    assert(exp);
-    if (snprintf(exp, len, "-(%s)", e->expression) < 0) {
-        assert(0);
-    }
-    free(e->expression);
-    e->expression = exp;
+    display_surround("-(", ")", 0); // FIXME: Cursor
 }
 
 
 static void
 exp_inv(void)
 {
-    struct exprm_state *e = get_state();
-
-    /* Ending zero + 1/ + parenthesis */
-    int len = strlen(e->expression) + 5;
-    char *exp = malloc(len);
-
-    assert(exp);
-    if (snprintf(exp, len, "1/(%s)", e->expression) < 0) {
-        assert(0);
-    }
-    free(e->expression);
-    e->expression = exp;
-}
-
-
-char *
-str_replace(char *str, char *from, char *to)
-{
-    char output[MAXLINE];
-    int offset = 0;
-    char *c;
-    int flen = strlen(from);
-    int tlen = strlen(to);
-    
-    for (c = str; *c && offset < MAXLINE - 1; c++, offset++) {
-        if (strncasecmp(from, c, flen) == 0) {
-            SNPRINTF(output + offset, MAXLINE - offset, to);
-            c += flen - 1;
-            offset += tlen - 1;
-        } else {
-            output[offset] = *c;
-        }
-    }
-
-    if (offset >= MAXLINE)
-        offset = MAXLINE - 1;
-    output[offset] = '\0';
-    
-    return strdup(output);
+    display_surround("1/(", ")", 0); // FIXME: Cursor
 }
 
 
@@ -462,29 +162,26 @@ void
 do_expression(int function, int arg, int cursor)
 {
     char buf[MAXLINE];
-    struct exprm_state *e, *ex;
+    struct exprm_state *e;
     
-    new_state();
+    display_push();
     e = get_state();
-    ex = peek_previous_state();
 
     e->cursor = cursor;
-    e->clear = 0;
 
     ui_set_statusbar("", "");
 
     /* Starting a number after a calculation clears the display */
-    if (strcmp(e->expression, "Ans") == 0) {
+    if (display_is_result()) {
         if (buttons[function].flags & NUMBER) {
-            exp_replace("");
+            display_set_string("");
         }
     }
 
     switch (buttons[function].id) {
         case KEY_CLEAR:
         case KEY_CLEAR_ENTRY:
-            e->clear = 1;
-            exp_clear();
+            display_clear(FALSE);
             ui_set_error_state(FALSE);
             MPstr_to_num("0", DEC, e->ans);
             break;
@@ -512,19 +209,20 @@ do_expression(int function, int arg, int cursor)
 
         case KEY_RECALL:
             SNPRINTF(buf, MAXLINE, "R%d", arg);
-            cursor = exp_insert(buf, cursor);
+            cursor = display_insert(buf, cursor);
             break;
 
         case KEY_CONSTANT:
-            cursor = exp_insert(make_number(v->MPcon_vals[arg], v->base, FALSE), cursor);
+            make_number(buf, MAXLINE, v->MPcon_vals[arg], v->base, FALSE);
+            cursor = display_insert(buf, cursor);
             break;
 
         case KEY_BACKSPACE:
-            cursor = exp_backspace(cursor);
+            cursor = display_backspace(cursor);
             break;
         
         case KEY_DELETE:
-            cursor = exp_delete(cursor);
+            cursor = display_delete(cursor);
             break;
 
         case KEY_CHANGE_SIGN:
@@ -542,7 +240,7 @@ do_expression(int function, int arg, int cursor)
              * this result */
             /* TODO: Work out why two undo steps are required and why
              * the cursor must be taken from the first undo */
-            if (strcmp(e->expression, "Ans") == 0) {
+            if (display_is_result()) {
                 perform_undo();
                 e = get_state();
                 cursor = e->cursor;
@@ -551,35 +249,20 @@ do_expression(int function, int arg, int cursor)
                 }
 
             /* Do nothing */                
-            } else if (e->expression[0] == '\0') {
+            } else if (display_is_empty()) {
                 ;
                 
             /* Solve the equation */
             } else {
                 int MPval[MP_SIZE];
-                char *c, *message = NULL;
-                GString *clean;
                 int result;
+                const char *message = NULL;
                 
-                /* Remove thousands separators and use english radix */
-                clean = g_string_sized_new(strlen(e->expression));
-                for (c = e->expression; *c; c++) {
-                    if (v->tsep[0] != '\0' && strncmp(c, v->tsep, strlen(v->tsep)) == 0) {
-                        c += strlen(v->tsep) - 1;
-                    } else if (strncmp(c, v->radix, strlen(v->radix)) == 0) {
-                        g_string_append_c(clean, '.');
-                        c += strlen(v->radix) - 1;
-                    } else {
-                        g_string_append_c(clean, *c);
-                    }
-                }
-                result = ce_parse(clean->str, MPval);
-                g_string_free(clean, TRUE);
-                
+                result = display_solve(MPval);
                 switch (result) {
                     case 0:
                         mpstr(MPval, e->ans);
-                        exp_replace("Ans");
+                        display_set_string("Ans");
                         cursor = -1;
                         break;
 
@@ -613,38 +296,25 @@ do_expression(int function, int arg, int cursor)
             break;
 
         case KEY_NUMERIC_POINT:
-            cursor = exp_insert(ui_get_localized_numeric_point(), cursor);
+            cursor = display_insert(v->radix, cursor);
             break;
 
         default:
             /* If display is a number then perform functions on that number */
-            if ((buttons[function].flags & (PREFIXOP | FUNC))
-                && !strcmp(e->expression, "Ans")
-                && !(ex && ex->clear)) {
-                SNPRINTF(buf, MAXLINE, "%s(%s)",
-                         buttons[function].symname,
-                         e->expression);
-                exp_replace(buf);
-            } else if ((((buttons[function].id == KEY_ADD) 
-                         || (buttons[function].id == KEY_SUBTRACT))
-                        && !strcmp(e->expression, "Ans")
-                        && (ex && ex->clear))) {
-                exp_replace(buttons[function].symname);
+            if (buttons[function].flags & (PREFIXOP | FUNC) && display_is_result()) {
+                SNPRINTF(buf, MAXLINE, "%s(", buttons[function].symname);
+                display_surround(buf, ")", 0); // FIXME: Cursor
             } else {
                 if (buttons[function].flags & FUNC) {
                     SNPRINTF(buf, MAXLINE, "%s(", buttons[function].symname);
-                    if (!ex || ex->clear) {
-                        exp_replace(buf);
-                    } else {
-                        cursor = exp_insert(buf, cursor);
-                    }
+                    cursor = display_insert(buf, cursor);
                 } else {
-                    cursor = exp_insert(buttons[function].symname, cursor);
+                    cursor = display_insert(buttons[function].symname, cursor);
                 }
             }
             break;
     }
-    refresh_display(cursor);
+    display_refresh(cursor);
 }
 
 
@@ -655,8 +325,8 @@ do_calc(void)      /* Perform arithmetic calculation and display result. */
     int MP1[MP_SIZE], MP2[MP_SIZE];
 
     if (v->current == KEY_CALCULATE &&
-        v->old_cal_value == KEY_CALCULATE) {
-        if (v->new_input) {
+        v->ltr.old_cal_value == KEY_CALCULATE) {
+        if (v->ltr.new_input) {
             mpstr(v->MPlast_input, v->MPresult);
         } else {
             mpstr(v->MPlast_input, v->MPdisp_val);
@@ -664,11 +334,11 @@ do_calc(void)      /* Perform arithmetic calculation and display result. */
     }
 
     if (v->current != KEY_CALCULATE &&
-        v->old_cal_value == KEY_CALCULATE) {
-        v->cur_op = -1;
+        v->ltr.old_cal_value == KEY_CALCULATE) {
+        v->ltr.cur_op = -1;
     }
 
-    switch (v->cur_op) {
+    switch (v->ltr.cur_op) {
         case KEY_SIN:
         case KEY_SINH:
         case KEY_ASIN:
@@ -706,17 +376,15 @@ do_calc(void)      /* Perform arithmetic calculation and display result. */
                 ui_set_statusbar(_("Error, operands must be integers"),
                                  "gtk-dialog-error");
             } else {
-                int val;
-
                 mpdiv(v->MPresult, v->MPdisp_val, MP1);
                 mpcmim(MP1, MP1);
                 mpmul(MP1, v->MPdisp_val, MP2);
                 mpsub(v->MPresult, MP2, v->MPresult);
 
-                val = 0;
-                mpcim(&val, MP1);
-                if ((mplt(v->MPdisp_val, MP1) && mpgt(v->MPresult, MP1)) ||
-                    mplt(v->MPresult, MP1)) { 
+                mp_set_from_integer(0, MP1);
+                if ((mp_is_less_than(v->MPdisp_val, MP1)
+		     && mp_is_greater_than(v->MPresult, MP1)) ||
+                    mp_is_less_than(v->MPresult, MP1)) { 
                     mpadd(v->MPresult, v->MPdisp_val, v->MPresult);
                 }
             }
@@ -727,49 +395,49 @@ do_calc(void)      /* Perform arithmetic calculation and display result. */
             break;
 
         case KEY_AND:
-            mpcmd(v->MPresult, &dres);
-            mpcmd(v->MPdisp_val, &dval);
+            dres = mp_cast_to_double(v->MPresult);
+            dval = mp_cast_to_double(v->MPdisp_val);
             dres = setbool(ibool(dres) & ibool(dval));
-            mpcdm(&dres, v->MPresult);
+            mp_set_from_double(dres, v->MPresult);
             break;
 
         case KEY_OR:
-            mpcmd(v->MPresult, &dres);
-            mpcmd(v->MPdisp_val, &dval);
+            dres = mp_cast_to_double(v->MPresult);
+            dval = mp_cast_to_double(v->MPdisp_val);
             dres = setbool(ibool(dres) | ibool(dval));
-            mpcdm(&dres, v->MPresult);
+            mp_set_from_double(dres, v->MPresult);
             break;
 
         case KEY_XOR:
-            mpcmd(v->MPresult, &dres);
-            mpcmd(v->MPdisp_val, &dval);
+            dres = mp_cast_to_double(v->MPresult);
+            dval = mp_cast_to_double(v->MPdisp_val);
             dres = setbool(ibool(dres) ^ ibool(dval));
-            mpcdm(&dres, v->MPresult);
+            mp_set_from_double(dres, v->MPresult);
             break;
 
         case KEY_XNOR:
-            mpcmd(v->MPresult, &dres);
-            mpcmd(v->MPdisp_val, &dval);
+            dres = mp_cast_to_double(v->MPresult);
+            dval = mp_cast_to_double(v->MPdisp_val);
             dres = setbool(~ibool(dres) ^ ibool(dval));
-            mpcdm(&dres, v->MPresult);
+            mp_set_from_double(dres, v->MPresult);
 
         default:
             break;
     }
 
-    show_display(v->MPresult);
+    display_set_number(v->MPresult);
 
     if (!(v->current == KEY_CALCULATE &&
-          v->old_cal_value == KEY_CALCULATE)) {
+          v->ltr.old_cal_value == KEY_CALCULATE)) {
         mpstr(v->MPdisp_val, v->MPlast_input);
     }
 
     mpstr(v->MPresult, v->MPdisp_val);
     if (v->current != KEY_CALCULATE) {
-        v->cur_op = v->current;
+        v->ltr.cur_op = v->current;
     }
-    v->old_cal_value = v->current;
-    v->new_input     = v->key_exp = 0;
+    v->ltr.old_cal_value = v->current;
+    v->ltr.new_input     = v->ltr.key_exp = 0;
 }
 
 
@@ -777,7 +445,7 @@ void
 do_sin(void)
 {
     calc_trigfunc(sin_t, v->MPdisp_val, v->MPresult);
-    show_display(v->MPresult);
+    display_set_number(v->MPresult);
     mpstr(v->MPresult, v->MPdisp_val);
 }
 
@@ -786,7 +454,7 @@ void
 do_sinh(void)
 {
     calc_trigfunc(sinh_t, v->MPdisp_val, v->MPresult);
-    show_display(v->MPresult);
+    display_set_number(v->MPresult);
     mpstr(v->MPresult, v->MPdisp_val);
 }
 
@@ -795,7 +463,7 @@ void
 do_asin(void)
 {
     calc_trigfunc(asin_t, v->MPdisp_val, v->MPresult);
-    show_display(v->MPresult);
+    display_set_number(v->MPresult);
     mpstr(v->MPresult, v->MPdisp_val);
 }
 
@@ -804,7 +472,7 @@ void
 do_asinh(void)
 {
     calc_trigfunc(asinh_t, v->MPdisp_val, v->MPresult);
-    show_display(v->MPresult);
+    display_set_number(v->MPresult);
     mpstr(v->MPresult, v->MPdisp_val);
 }
 
@@ -813,7 +481,7 @@ void
 do_cos(void)
 {
     calc_trigfunc(cos_t, v->MPdisp_val, v->MPresult);
-    show_display(v->MPresult);
+    display_set_number(v->MPresult);
     mpstr(v->MPresult, v->MPdisp_val);
 }
 
@@ -822,7 +490,7 @@ void
 do_cosh(void)
 {
     calc_trigfunc(cosh_t, v->MPdisp_val, v->MPresult);
-    show_display(v->MPresult);
+    display_set_number(v->MPresult);
     mpstr(v->MPresult, v->MPdisp_val);
 }
 
@@ -831,7 +499,7 @@ void
 do_acos(void)
 {
     calc_trigfunc(acos_t, v->MPdisp_val, v->MPresult);
-    show_display(v->MPresult);
+    display_set_number(v->MPresult);
     mpstr(v->MPresult, v->MPdisp_val);
 }
 
@@ -840,7 +508,7 @@ void
 do_acosh(void)
 {
     calc_trigfunc(acosh_t, v->MPdisp_val, v->MPresult);
-    show_display(v->MPresult);
+    display_set_number(v->MPresult);
     mpstr(v->MPresult, v->MPdisp_val);
 }
 
@@ -849,7 +517,7 @@ void
 do_tan(void)
 {
     calc_trigfunc(tan_t, v->MPdisp_val, v->MPresult);
-    show_display(v->MPresult);
+    display_set_number(v->MPresult);
     mpstr(v->MPresult, v->MPdisp_val);
 }
 
@@ -858,7 +526,7 @@ void
 do_tanh(void)
 {
     calc_trigfunc(tanh_t, v->MPdisp_val, v->MPresult);
-    show_display(v->MPresult);
+    display_set_number(v->MPresult);
     mpstr(v->MPresult, v->MPdisp_val);
 }
 
@@ -867,7 +535,7 @@ void
 do_atan(void)
 {
     calc_trigfunc(atan_t, v->MPdisp_val, v->MPresult);
-    show_display(v->MPresult);
+    display_set_number(v->MPresult);
     mpstr(v->MPresult, v->MPdisp_val);
 }
 
@@ -876,7 +544,7 @@ void
 do_atanh(void)
 {
     calc_trigfunc(atanh_t, v->MPdisp_val, v->MPresult);
-    show_display(v->MPresult);
+    display_set_number(v->MPresult);
     mpstr(v->MPresult, v->MPdisp_val);
 }
 
@@ -885,7 +553,7 @@ void
 do_percent(void)
 {
     calc_percent(v->MPdisp_val, v->MPresult);
-    show_display(v->MPresult);
+    display_set_number(v->MPresult);
     mpstr(v->MPresult, v->MPdisp_val);
 }
 
@@ -894,11 +562,11 @@ do_percent(void)
 void
 do_clear(void)
 {
-    clear_display(TRUE);
+    display_clear(TRUE);
     if (v->error) {
         ui_set_display("", -1);
     }
-    initialise();
+    display_reset();
 }
 
 
@@ -906,7 +574,7 @@ do_clear(void)
 void
 do_clear_entry(void)
 {
-    clear_display(FALSE);
+    display_clear(FALSE);
 }
 
 
@@ -927,14 +595,14 @@ do_base(enum base_type b)
 
         case EXPRS:
             e = get_state();
-            ret = usable_num(MP);
+            ret = display_is_usable_number(MP);
 
             if (ret) {
                 ui_set_statusbar(_("No sane value to convert"),
                                  "gtk-dialog-error");
             } else {
                 mpstr(MP, e->ans);
-                exp_replace("Ans");
+                display_set_string("Ans");
             }
             v->base = b;
             set_resource(R_BASE, Rbstr[(int) v->base]);
@@ -948,7 +616,7 @@ do_base(enum base_type b)
             break;
     }
 
-    refresh_display(-1);
+    display_refresh(-1);
 }
 
 
@@ -962,7 +630,7 @@ do_constant(int index)
 
     MPval = v->MPcon_vals[index];
     mpstr(MPval, v->MPdisp_val);
-    v->new_input = 1;
+    v->ltr.new_input = 1;
     syntaxdep_show_display();
 }
 
@@ -982,23 +650,22 @@ do_backspace(void)
  *  the exponent sign, then this reverts to entering a fixed point number.
  */
 
-    if (v->key_exp && !(strchr(v->display, '+'))) {
-        v->key_exp = 0;
+    if (v->ltr.key_exp && !(strchr(v->display, '+'))) {
+        v->ltr.key_exp = 0;
         v->display[strlen(v->display)-1] = '\0';
     }
 
 /* If we've backspaced over the numeric point, clear the pointed flag. */
 
-    if (v->pointed && !(strchr(v->display, '.'))) {
-        v->pointed = 0;
+    if (v->ltr.pointed && !(strchr(v->display, '.'))) {
+        v->ltr.pointed = 0;
     }
 
-    ui_set_display(v->display, -1);
+    display_set_string(v->display);
     MPstr_to_num(v->display, v->base, v->MPdisp_val);
 
     if (v->dtype == FIX) {
-        STRNCPY(v->fnum, v->display, MAX_LOCALIZED - 1);
-        ui_set_display(v->fnum, -1);
+        display_set_string(v->display);
     }
 }
 
@@ -1029,15 +696,15 @@ do_exchange(int index)
         case EXPRS:
             e = get_state();
 
-            if (usable_num(MPexpr)) {
+            if (display_is_usable_number(MPexpr)) {
                 ui_set_statusbar(_("No sane value to store"),
                                  "gtk-dialog-error");
             } else {
                 mpstr(v->MPmvals[index], MPtemp);
                 mpstr(MPexpr, v->MPmvals[index]);
                 mpstr(MPtemp, e->ans);
-                exp_replace("Ans");
-                refresh_display(-1);
+                display_set_string("Ans");
+                display_refresh(-1);
                 ui_make_registers();
             }
             break;
@@ -1046,7 +713,7 @@ do_exchange(int index)
             assert(0);
     }
 
-    v->new_input = 1;
+    v->ltr.new_input = 1;
     syntaxdep_show_display();
 }
 
@@ -1055,20 +722,20 @@ do_exchange(int index)
 void
 do_expno(void)
 {
-    v->pointed = (strchr(v->display, '.') != NULL);
-    if (!v->new_input) {
+    v->ltr.pointed = (strchr(v->display, '.') != NULL);
+    if (!v->ltr.new_input) {
         STRNCPY(v->display, "1.0 +", MAX_LOCALIZED - 1);
-        v->new_input = v->pointed = 1;
-    } else if (!v->pointed) {
+        v->ltr.new_input = v->ltr.pointed = 1;
+    } else if (!v->ltr.pointed) {
         STRNCAT(v->display, ". +", 3);
-        v->pointed = 1;
-    } else if (!v->key_exp) {
+        v->ltr.pointed = 1;
+    } else if (!v->ltr.key_exp) {
         STRNCAT(v->display, " +", 2);
     }
-    v->toclear = 0;
-    v->key_exp = 1;
-    v->exp_posn = strchr(v->display, '+');
-    ui_set_display(v->display, -1);
+    v->ltr.toclear = 0;
+    v->ltr.key_exp = 1;
+    v->ltr.exp_posn = strchr(v->display, '+');
+    display_set_string(v->display);
     MPstr_to_num(v->display, v->base, v->MPdisp_val);
 }
 
@@ -1092,22 +759,21 @@ do_factorial(int *MPval, int *MPres)
 
     mpstr(MPval, MPa);
     mpcmim(MPval, MP1);
-    i = 0;
-    mpcim(&i, MP2);
-    if (mpeq(MPval, MP1) && mpge(MPval, MP2)) {   /* Only positive integers. */
-        i = 1;
-        if (mpeq(MP1, MP2)) {                     /* Special case for 0! */
-            mpcim(&i, MPres);
+    mp_set_from_integer(0, MP2);
+    if (mp_is_equal(MPval, MP1)
+	&& mp_is_equal(MPval, MP2)) {   /* Only positive integers. */
+        if (mp_is_equal(MP1, MP2)) {    /* Special case for 0! */
+            mp_set_from_integer(1, MPres);
             return;
         }
-        mpcim(&i, MPa);
-        mpcmi(MP1, &i);
+        mp_set_from_integer(1, MPa);
+        i = mp_cast_to_int(MP1);
         if (!i) {
             matherr((struct exception *) NULL);
         } else {
             while (i > 0) {
                 mpmuli(MPa, &i, MPa);
-                mpcmd(MPa, &val);
+                val = mp_cast_to_double(MPa);
                 if (v->error) {
                     mperr();
                     return;
@@ -1153,7 +819,7 @@ do_function(int index)      /* Perform a user defined function. */
     } else {
         ui_set_statusbar(_("Malformed function"), "gtk-dialog-error");
     }
-    v->new_input = 1;
+    v->ltr.new_input = 1;
 }
 
 
@@ -1221,15 +887,15 @@ do_immedfunc(int s[MP_SIZE], int t[MP_SIZE])
             break;
 
         case KEY_CHANGE_SIGN:
-            if (v->key_exp) {
-                if (*v->exp_posn == '+') {
-                    *v->exp_posn = '-';
+            if (v->ltr.key_exp) {
+                if (*v->ltr.exp_posn == '+') {
+                    *v->ltr.exp_posn = '-';
                 } else {
-                    *v->exp_posn = '+';
+                    *v->ltr.exp_posn = '+';
                 }
-                ui_set_display(v->display, -1);
+                display_set_string(v->display);
                 MPstr_to_num(v->display, v->base, s);
-                v->key_exp = 0;
+                v->ltr.key_exp = 0;
             } else {
                 mpneg(s, t);
             }
@@ -1242,7 +908,7 @@ void
 do_immed(void)
 {
     do_immedfunc(v->MPdisp_val, v->MPdisp_val);
-    show_display(v->MPdisp_val);
+    display_set_number(v->MPdisp_val);
 }
 
 
@@ -1251,9 +917,9 @@ do_number(void)
 {
     int offset;
 
-    if (v->toclear) {
+    if (v->ltr.toclear) {
         offset = 0;
-        v->toclear = 0;
+        v->ltr.toclear = 0;
     } else {
         offset = strlen(v->display);
     }
@@ -1261,9 +927,9 @@ do_number(void)
         SNPRINTF(v->display+offset, MAXLINE-offset, "%s", buttons[v->current].symname);
     }
 
-    ui_set_display(v->display, -1);
+    display_set_string(v->display);
     MPstr_to_num(v->display, v->base, v->MPdisp_val);
-    v->new_input = 1;
+    v->ltr.new_input = 1;
 }
 
 
@@ -1283,13 +949,13 @@ do_numtype(enum num_type n)   /* Set number display type. */
 
         case EXPRS:
             e = get_state();
-            ret = usable_num(MP);
+            ret = display_is_usable_number(MP);
             if (ret) {
                 ui_set_statusbar(_("No sane value to convert"),
                                  "gtk-dialog-error");
             } else {
                 mpstr(MP, e->ans);
-                exp_replace("Ans");
+                display_set_string("Ans");
                 ui_make_registers();
             }
             clear_undo_history();
@@ -1299,7 +965,7 @@ do_numtype(enum num_type n)   /* Set number display type. */
         assert(0);
     }
 
-    refresh_display(-1);
+    display_refresh(-1);
 }
 
 
@@ -1310,20 +976,20 @@ do_paren(void)
 
     switch (v->current) {
         case KEY_START_BLOCK:
-            if (v->noparens == 0) {
-                if (v->cur_op == -1) {
+            if (v->ltr.noparens == 0) {
+                if (v->ltr.cur_op == -1) {
                     v->display[0] = 0;
                     ui_set_statusbar(_("Cleared display, prefix without an operator is not allowed"), "");
                 } else {
-                    paren_disp(v->cur_op);
+                    paren_disp(v->ltr.cur_op);
                 }
             }
-            v->noparens++;
+            v->ltr.noparens++;
             break;
 
         case KEY_END_BLOCK:
-            v->noparens--;
-            if (!v->noparens) {
+            v->ltr.noparens--;
+            if (!v->ltr.noparens) {
                 int ret, i = 0;
                 while (v->display[i++] != '(') {
                     /* do nothing */;
@@ -1331,7 +997,7 @@ do_paren(void)
 
                 ret = lr_parse(&v->display[i], v->MPdisp_val);
                 if (!ret) {
-                    show_display(v->MPdisp_val);
+                    display_set_number(v->MPdisp_val);
                     return;
                 } else {
                     ui_set_statusbar(_("Malformed parenthesis expression"),
@@ -1357,7 +1023,7 @@ do_sto(int index)
             break;
 
         case EXPRS:
-            if (usable_num(v->MPmvals[index])) {
+            if (display_is_usable_number(v->MPmvals[index])) {
                 ui_set_statusbar(_("No sane value to store"),
                                  "gtk-dialog-error");
             }
@@ -1392,7 +1058,7 @@ void
 do_rcl(int index)
 {
     mpstr(v->MPmvals[index], v->MPdisp_val);
-    v->new_input = 1;
+    v->ltr.new_input = 1;
     syntaxdep_show_display();
 }
 
@@ -1419,11 +1085,11 @@ syntaxdep_show_display(void)
 {
     switch (v->syntax) {
         case NPA:
-            show_display(v->MPdisp_val);
+            display_set_number(v->MPdisp_val);
             break;
 
         case EXPRS:
-     	    refresh_display(-1);
+     	    display_refresh(-1);
             break;
 
         default:
@@ -1435,16 +1101,16 @@ syntaxdep_show_display(void)
 void
 do_point(void)                   /* Handle numeric point. */
 {
-    if (!v->pointed) {
-        if (v->toclear) {
+    if (!v->ltr.pointed) {
+        if (v->ltr.toclear) {
             STRNCPY(v->display, ".", MAX_LOCALIZED - 1);
-            v->toclear = 0;
+            v->ltr.toclear = 0;
         } else {
             STRCAT(v->display, ".");
         }
-        v->pointed = 1;
+        v->ltr.pointed = 1;
     }
-    ui_set_display(v->display, -1);
+    display_set_string(v->display);
     MPstr_to_num(v->display, v->base, v->MPdisp_val);
 }
 
@@ -1477,7 +1143,7 @@ void
 do_portion(void)
 {
     do_portionfunc(v->MPdisp_val);
-    show_display(v->MPdisp_val);
+    display_set_number(v->MPdisp_val);
 }
 
 
@@ -1492,7 +1158,7 @@ do_shift(int count)     /* Perform bitwise shift on display value. */
     switch (v->syntax) {
         case NPA:
             MPstr_to_num(v->display, v->base, MPtemp);
-            mpcmd(MPtemp, &dval);
+            dval = mp_cast_to_double(MPtemp);
             temp = ibool(dval);
 
             if (count < 0) {
@@ -1502,14 +1168,14 @@ do_shift(int count)     /* Perform bitwise shift on display value. */
             }
 
             dval = setbool(temp);
-            mpcdm(&dval, v->MPdisp_val);
-            show_display(v->MPdisp_val);
+            mp_set_from_double(dval, v->MPdisp_val);
+            display_set_number(v->MPdisp_val);
             mpstr(v->MPdisp_val, v->MPlast_input);
             break;
 
         case EXPRS:
             e = get_state();
-            if (usable_num(MPval) || !is_integer(MPval)) {
+            if (display_is_usable_number(MPval) || !is_integer(MPval)) {
                 ui_set_statusbar(_("No sane value to do bitwise shift"),
                                  "gtk-dialog-error");
                 break;
@@ -1517,47 +1183,13 @@ do_shift(int count)     /* Perform bitwise shift on display value. */
 
             calc_shift(MPval, e->ans, count);
 
-            exp_replace("Ans");
+            display_set_string("Ans");
             break;
 
         default:
             assert(0);
     }
 
-    v->new_input = 1;
+    v->ltr.new_input = 1;
     syntaxdep_show_display();
 }
-
-
-void
-do_trigtype(enum trig_type t)    /* Change the current trigonometric type. */
-{
-    v->ttype = t;
-    set_resource(R_TRIG, Rtstr[(int) v->ttype]);
-    switch (v->cur_op) {
-        case KEY_SIN:
-        case KEY_SINH:
-        case KEY_ASIN:
-        case KEY_ASINH:
-        case KEY_COS:
-        case KEY_COSH:
-        case KEY_ACOS:
-        case KEY_ACOSH:
-        case KEY_TAN:
-        case KEY_TANH:
-        case KEY_ATAN:
-        case KEY_ATANH:
-            mpstr(v->MPtresults[(int) v->ttype], v->MPdisp_val);
-            show_display(v->MPtresults[(int) v->ttype]);
-            break;
-
-        default:
-            break;
-    }
-}
-
-void
-show_error(char *message)
-{
-    ui_set_statusbar(message, "gtk-dialog-error");
-}
